/***************************************************************************

	cdda.c

	NEOGEO CD CD-DA emulation

***************************************************************************/

#include "ncdz.h"


/***************************************************************************
	Global variables
 **************************************************************************/

int	neogeo_cdda_enable;
int	neogeo_cdda_volume;
int cdda_current_track = 0;
volatile int cdda_playing = 0;
volatile int cdda_autoloop = 0;


/***************************************************************************
	Function prototypes
 **************************************************************************/

static void neogeo_cdda_command(int command, int track);


/*------------------------------------------------------
	Initialize CDDA emulation
 -----------------------------------------------------*/

int	cdda_init(void)
{
	cdda_current_track = 0;
	cdda_playing = CDDA_STOP;

	return mp3_thread_start();
}


/*------------------------------------------------------
	Shutdown CDDA emulation
 -----------------------------------------------------*/

void cdda_shutdown(void)
{
	mp3_thread_stop();
}


/*------------------------------------------------------
	CDDA play
 -----------------------------------------------------*/

void cdda_play(int track)
{
	cdda_current_track = track;
	cdda_playing = CDDA_PLAY;

	if (neogeo_cdda_enable)
	{
		char path[MAX_PATH], pattern[16], *fname;

		strcpy(path, mp3_dir);
		sprintf(pattern, "%02d.mp3", track);

		if ((fname = find_file(pattern, path)))
		{
			strcat(path, fname);
			mp3_play(path);
		}
	}
}


/*------------------------------------------------------
	CDDA stop
 -----------------------------------------------------*/

void cdda_stop(void)
{
	cdda_playing = CDDA_STOP;
	cdda_current_track = 0;

	if (neogeo_cdda_enable) mp3_stop();
}


/*------------------------------------------------------
	CDDA pause
 -----------------------------------------------------*/

void cdda_pause(void)
{
	cdda_playing = CDDA_PAUSE;

	if (neogeo_cdda_enable) mp3_pause(1);
}


/*------------------------------------------------------
	CDDA resume
 -----------------------------------------------------*/

void cdda_resume(void)
{
	cdda_playing = CDDA_PLAY;

	if (neogeo_cdda_enable) mp3_pause(0);
}


/*------------------------------------------------------
	CDDA control (BIOS function)
 -----------------------------------------------------*/

void neogeo_cdda_control(void)
{
	int command, track;
	u32 offset;

	command = m68000_get_reg(M68K_D0);
	track   = command & 0xff;
	command = (command >> 8) & 0xff;

	if ((command & 2) == 0)
	{
		m68k_write_memory_8(0x108000 + 0x764b, track);
		m68k_write_memory_8(0x108000 + 0x76f8, track);
		m68k_write_memory_8(0x108000 + 0x76f7, command);
	}
	m68k_write_memory_8(0x108000 + 0x76f6, command);

	if (command <= 7)
	{
		if (neogeo_cdda_enable && (command || track))
			neogeo_cdda_command(command, track);
	}

	offset = m68k_read_memory_32(0x108000 + 0x76ea);

	if (offset)
	{
		m68k_write_memory_8(0x108000 + 0x7678, 0x01);

		if (command <= 7)
		{
			offset  = (offset - 0xe00000) >> 1;
			neogeo_z80_memory[offset + 0] = 0;
			neogeo_z80_memory[offset + 1] = 0;
		}
	}
}


/*------------------------------------------------------
	CDDA check command and track
 -----------------------------------------------------*/

void neogeo_cdda_check(void)
{
	u32 offset = m68k_read_memory_32(0x108000 + 0x76ea);

	if (offset)
	{
		u8 command, track;

		offset  = (offset - 0xe00000) >> 1;

		command = neogeo_z80_memory[offset + 0];
		track   = neogeo_z80_memory[offset + 1];

		neogeo_z80_memory[offset + 0] = 0;
		neogeo_z80_memory[offset + 1] = 0;

		if (command <= 7)
		{
			if (neogeo_cdda_enable && (command || track))
				neogeo_cdda_command(command, track);
		}
	}
}


/*------------------------------------------------------
	Send CDDA command

	R}h̒l
	bit 1: I[g[vtO (0: 1:L)
	bit 2: vC@ (0:Đ 1:|[Y/W[)
	bit 3: R}ht҂ (0:҂Ȃ 1:҂)
     bit 3͏LƎvB͂Ă܂B
 -----------------------------------------------------*/

static void neogeo_cdda_command(int command, int track)
{
	int loop, _track;

	_track = track;
	track = ((track >> 4) * 10) + (track & 0x0f);
	loop  = !(command & 1);

	switch (command)
	{
	case 0:
	case 1:
	case 4:
	case 5:
		if (track > 1 && track <= 99)
		{
			cdda_autoloop = loop;
			cdda_play(track);
		}
		break;

	case 2:
	case 6:
		if (cdda_playing == CDDA_PLAY)
		{
			cdda_pause();
		}
		break;

	case 3:
	case 7:
		if (cdda_playing == CDDA_PAUSE)
		{
			cdda_resume();
		}
		break;
	}
}


#ifdef SAVE_STATE

/*------------------------------------------------------
	Save CDDA state
 -----------------------------------------------------*/

void state_save_cdda(u8 *state_buf)
{
	u8 *in = state_buf;
	int playing = cdda_playing;
	int autoloop = cdda_autoloop;
	u32 frame = mp3_get_current_frame();

	state_save_long(&cdda_current_track, 1);
	state_save_long(&playing,  1);
	state_save_long(&autoloop, 1);
	state_save_long(&frame, 1);
}


/*------------------------------------------------------
	Load CDDA state
 -----------------------------------------------------*/

void state_load_cdda(u8 *state_buf)
{
	if (neogeo_cdda_enable)
	{
		u8 *out = state_buf;
		int playing;
		int autoloop;
		u32 frame;

		state_load_long(&cdda_current_track, 1);
		state_load_long(&playing, 1);
		state_load_long(&autoloop, 1);
		state_load_long(&frame, 1);

		cdda_playing = playing;
		cdda_autoloop = autoloop;

		if (cdda_playing)
		{
			char path[MAX_PATH], pattern[16], *fname;

			strcpy(path, mp3_dir);
			sprintf(pattern, "%02d.mp3", cdda_current_track);

			if ((fname = find_file(pattern, path)))
			{
				strcat(path, fname);
				mp3_seek_set(path, frame, cdda_playing == CDDA_PAUSE);
			}
			else mp3_stop();
		}
		else cdda_stop();
	}
}

#endif
