/***************************************************************************

	input.c

    PSP input control functions

***************************************************************************/

#include "psp.h"


/***************************************************************************
	Private variables
 ***************************************************************************/

static int pad;
static u8 pressed_check;
static u8 pressed_count;
static u8 pressed_delay;
static TICKER curr_time;
static TICKER prev_time;


/*--------------------------------------------------------
	Initialize pad
 -------------------------------------------------------*/

void pad_init(void)
{
	sceCtrlSetSamplingCycle(0);
	sceCtrlSetSamplingMode(1);

	pad = 0;
	pressed_check = 0;
	pressed_count = 0;
	pressed_delay = 0;
}


/*--------------------------------------------------------
	Poll pad
 -------------------------------------------------------*/

int poll_gamepad(void)
{
	SceCtrlData paddata;

	sceCtrlPeekBufferPositive(&paddata, 1);

	if (paddata.Ly >= 0xD0) paddata.Buttons |= PSP_CTRL_DOWN;  // DOWN
	if (paddata.Ly <= 0x30) paddata.Buttons |= PSP_CTRL_UP;    // UP
	if (paddata.Lx <= 0x30) paddata.Buttons |= PSP_CTRL_LEFT;  // LEFT
	if (paddata.Lx >= 0xD0) paddata.Buttons |= PSP_CTRL_RIGHT; // RIGHT

	return paddata.Buttons;
}


/*--------------------------------------------------------
	Update pad status
 -------------------------------------------------------*/

void pad_update(void)
{
	u32 data;

	data = poll_gamepad() & PSP_CTRL_ANY;

	if (data)
	{
		if (!pressed_check)
		{
			pressed_check = 1;
			pressed_count = 0;
			pressed_delay = 8;
			prev_time = ticker();
		}
		else
		{
			int count;

			curr_time = ticker();
			count = (int)((curr_time - prev_time) / (TICKS_PER_SEC / 60));
			prev_time = curr_time;

			pressed_count += count;

			if (pressed_count > pressed_delay)
			{
				pressed_count = 0;
				if (pressed_delay > 2) pressed_delay -= 2;
			}
			else data = 0;
		}
	}
	else pressed_check = 0;

	pad = data;
}


/*--------------------------------------------------------
	Get button status
 -------------------------------------------------------*/

int pad_pressed(u32 code)
{
	return (pad & code) != 0;
}


/*--------------------------------------------------------
	Check any buttons are pressed
 -------------------------------------------------------*/

int pad_pressed_any(u32 disable_code)
{
	return (pad & (PSP_CTRL_ANY ^ disable_code)) != 0;
}


/*--------------------------------------------------------
	Wait clear pad status
 -------------------------------------------------------*/

void pad_wait_clear(void)
{
	while (poll_gamepad())
	{
		if (!bLoop) break;
	}

	pad = 0;
	pressed_check = 0;
}


/*--------------------------------------------------------
	Wait press any buttons
 -------------------------------------------------------*/

void pad_wait_press(int msec)
{
	pad_wait_clear();

	if (msec == -1)
	{
		while (1)
		{
			video_wait_vsync();
			if (poll_gamepad()) break;
			if (!bLoop) break;
		}
	}
	else
	{
		TICKER curr = ticker();
		TICKER target, prev;

		target = curr + msec * (TICKS_PER_SEC / 1000);

		while (curr - target < 0)
		{
			prev = curr;
			curr = ticker();

			video_wait_vsync();
			if (poll_gamepad()) break;
			if (prev > curr) break;
			if (!bLoop) break;
		}
	}

	pad_wait_clear();
}
